/* Copyright (c) 2016 Raffaele Arecchi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package aktrs.rt;

import java.io.IOException;
import java.io.Serializable;
import java.net.InetAddress;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import aktrs.messages.ReinitializeMessage;
import aktrs.messages.Stop;
import aktrs.util.AktrsFormatter;
import aktrs.util.ConsoleOutputHandler;

/**
 * The class containing actors and which is responsible of their life-cycle.<br>
 * <br>
 * An ActorSystem is capable to create an actor by
 * {@link #summon(String, Class)} or killing an actor by
 * {@link #dissolve(Actor)} <br>
 * An ActorSystem can optionally by linked to other ActorSystems via TCP
 * protocol. Actors are local or remote relative to an ActorSystem (see
 * {@link Actor}). If two ActorSystem are linked, all their actors are able to
 * talk each other.
 */
public class ActorSystem {
	public static final String SYSTEM_ACTOR_NAME = "system";

	public final Logger logger;
	public final String id;
	private Actor systemActor;
	private Map<ActorAddress, Actor> actorsRegistry;
	private Map<String, RemoteSystem> linkedSystems;
	private TcpListener tcpListener;

	/**
	 * Instantiate an ActorSystem with a custom logger
	 * 
	 * @param id
	 *            id of the ActorSystem
	 * @param logger
	 *            the custom Logger
	 */
	public ActorSystem(String id, Logger logger) {
		this.id = id;
		if (logger == null) {
			this.logger = Logger.getLogger(id);
			ConsoleOutputHandler handler = new ConsoleOutputHandler();
			handler.setFormatter(new AktrsFormatter());
			this.logger.addHandler(handler);
			setLogLevel(Level.INFO);
		} else {
			this.logger = logger;
		}
		actorsRegistry = new ConcurrentHashMap<ActorAddress, Actor>();
		linkedSystems = new ConcurrentHashMap<String, RemoteSystem>();
		systemActor = summon("/" + id + "/" + SYSTEM_ACTOR_NAME, SystemActor.class, null);
		actorsRegistry.put(systemActor.address, systemActor);
		this.logger.info("Actor System " + id + " started.");
	}

	/**
	 * Instantiate an ActorSystem with the default console logger
	 * 
	 * @param id
	 *            id of the ActorSystem
	 */
	public ActorSystem(String id) {
		this(id, null);
	}

	/**
	 * Instantiate an ActorSystem with the default console logger and which can
	 * be linked as a remote ActorSystem
	 * 
	 * @param id
	 *            id of the ActorSystem
	 * @param port
	 *            port for listening to incoming connections for the ActorSystem
	 */
	public ActorSystem(String id, int port) {
		this(id, null, port);
	}

	/**
	 * Instantiate an ActorSystem with a custom logger and which can be linked
	 * as a remote ActorSystem
	 * 
	 * @param id
	 *            id of the ActorSystem
	 * @param logger
	 *            the custom Logger
	 * @param port
	 *            port for listening to incoming connections for the ActorSystem
	 */
	public ActorSystem(String id, Logger logger, int port) {
		this(id, logger);
		tcpListener = new TcpListener(this, port);
		new Thread(tcpListener, "tcp-listener").start();
	}

	/**
	 * @return port for listening to incoming connections
	 */
	public int getListeningPort() {
		return tcpListener == null ? -1 : tcpListener.port;
	}

	/**
	 * @param address
	 *            the address of the actor in String format
	 * @return the actor located to the specified address, null if it does not
	 *         exist
	 */
	public Actor evoke(String address) {
		return evoke(new ActorAddress(address));
	}

	/**
	 * @param address
	 *            the address of the actor
	 * @return the actor located to the specified address, null if it does not
	 *         exist
	 */
	public Actor evoke(ActorAddress address) {
		return actorsRegistry.get(address);
	}

	/**
	 * Instantiate and run a new class-specific actor with the specified
	 * address. Calls {@link #summon(ActorAddress, Class, Object)}.
	 * 
	 * @param address
	 *            in String format
	 * @param klass
	 *            the class of the actor
	 * @return the summoned actor
	 */
	public Actor summon(String address, Class<? extends Actor> klass) {
		return summon(address, klass, null);
	}

	/**
	 * Instantiate and run a new class-specific actor with the specified address
	 * and the specified initialization object. Calls
	 * {@link #summon(ActorAddress, Class, Object)}.
	 * 
	 * @param address
	 *            in String format
	 * @param klass
	 *            the class of the actor
	 * @param initObj
	 *            the initialization object
	 * @return the summoned actor
	 */
	public Actor summon(String address, Class<? extends Actor> klass, Object initObj) {
		return summon(new ActorAddress(address), klass, initObj);
	}

	/**
	 * Instantiate and run a new class-specific actor with the specified
	 * address. Calls {@link #summon(ActorAddress, Class, Object)}.
	 * 
	 * @param address
	 *            the address of the actor
	 * @param klass
	 *            the class of the actor
	 * @return the summoned actor
	 */
	public Actor summon(ActorAddress address, Class<? extends Actor> klass) {
		return summon(address, klass, null);
	}

	/**
	 * Instantiate and run a new class-specific actor with the specified address
	 * and the specified initialization object. If an actor is already
	 * registered at the specified address, a re-initialization message is sent.
	 * If the address is relative to a linked ActorSystem, a remote actor is
	 * summoned on the remote ActorSystem and a proxy is created on this
	 * ActorSystem.
	 * 
	 * @param address
	 *            the address of the summoned actor
	 * @param klass
	 *            the class of the actor
	 * @param initObj
	 *            the initialization object, can be null.
	 * @return the summoned actor
	 * 
	 * @throws IllegalArgumentException
	 *             if 1) the actor is remote and the initialization object is
	 *             not serializable or 2) the address references a not-linked
	 *             ActorSystem
	 * @throws RuntimeException
	 *             when an error occurred summoning the actor relative to I/O or
	 *             class instantiation
	 */
	public Actor summon(ActorAddress address, Class<? extends Actor> klass, Object initObj) {
		if (logger.isLoggable(Level.FINE)) {
			logger.fine("Actor System " + id + ". Summoning  " + address.toString() + ", Class:  " + klass.getName() + ", Initialization object: " + initObj);
		}
		Actor actor;
		try {
			if (address.systemId.equals(id)) {
				if (logger.isLoggable(Level.FINE)) {
					logger.fine("Actor System " + id + ". Actor " + address.toString() + " is a local actor");
				}
				actor = actorsRegistry.get(address);
				if (actor != null) {
					if (logger.isLoggable(Level.FINE)) {
						logger.fine("Actor System " + id + ". Actor " + address.toString() + " exists.");
					}
					if (initObj != null) {
						if (logger.isLoggable(Level.FINE)) {
							logger.fine("Actor System " + id + ". Sending to Actor " + address.toString() + " a re-initializing message.");
						}
						systemActor.tell(actor, ReinitializeMessage.create(initObj));
					}
				} else {
					actor = klass.newInstance();
					actor.setSystem(this);
					actor.setAddress(address);
					actor.setInit(initObj);
					new Thread(actor, address.toString()).start();
					actorsRegistry.put(address, actor);
				}
			} else if (initObj != null && !(initObj instanceof Serializable)) {
				String notSummonMessage = "Actor System " + id + ". Cannot do a remote summon of " + address.toString() + ": init object must be serializable";
				logger.severe(notSummonMessage);
				throw new IllegalArgumentException(notSummonMessage);
			} else {
				if (logger.isLoggable(Level.FINE)) {
					logger.fine("Actor System " + id + ". Actor " + address.toString() + " is a remote actor");
				}
				actor = summonRemote(address, klass, (Serializable) initObj);
			}
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Actor System " + id + ". " + e.getMessage(), e);
			throw new RuntimeException(e);
		}
		if (logger.isLoggable(Level.FINE)) {
			logger.fine("Actor System " + id + ". Actor " + address.toString() + " summoned");
		}
		return actor;
	}

	private Actor summonRemote(ActorAddress address, Class<? extends Actor> klass, Serializable initObj) throws InstantiationException, IllegalAccessException {
		RemoteSystem remoteSystem = linkedSystems.get(address.systemId);
		if (remoteSystem == null) {
			String systemNotLinkedMessage = "Actor System " + id + ". System " + address.systemId + " is not linked";
			logger.severe(systemNotLinkedMessage);
			throw new IllegalArgumentException(systemNotLinkedMessage);
		}
		remoteSystem.summon(address, klass, initObj);
		Actor actor = klass.newInstance();
		actor.setAddress(address);
		actor.setRemoteSystem(remoteSystem);
		actor.setSystem(this);
		actorsRegistry.put(address, actor);
		return actor;
	}

	Actor mirrorRemoteActor(ActorAddress remoteAddress, Class<? extends Actor> remoteClass, RemoteSystem remoteSystem) throws InstantiationException,
			IllegalAccessException {
		Actor actor = remoteClass.newInstance();
		actor.setAddress(remoteAddress);
		actor.setRemoteSystem(remoteSystem);
		actor.setSystem(this);
		actorsRegistry.put(remoteAddress, actor);
		return actor;
	}

	RemoteSystem linkIfNotLinked(String remoteId, InetAddress netAddress, int port) {
		RemoteSystem remoteSys = linkedSystems.get(remoteId);
		if (remoteSys == null) {
			return linkAndGet(remoteId, netAddress, port);
		} else {
			if (logger.isLoggable(Level.FINE)) {
				logger.fine("Actor System " + id + ". Remote System " + remoteId + " already linked");
			}
			return remoteSys;
		}
	}

	RemoteSystem linkAndGet(String remoteId, InetAddress netAddress, int port) {
		logger.info("Actor System " + id + ". Linking Remote System " + remoteId);

		if (remoteId.equals(id)) {
			String cannotLink = "Actor System " + id + ". Cannot link the system: ID " + remoteId + " should be different";
			logger.severe(cannotLink);
			throw new IllegalArgumentException(cannotLink);
		}
		RemoteSystem remoteSys = linkedSystems.get(remoteId);
		if (remoteSys != null) {
			logger.info("Actor System " + id + ". Remote System " + remoteId + " has to be reconfigured, disconnecting it");
			remoteSys.disconnect();
		}
		RemoteSystem updatedRemoteSys = new RemoteSystem(this, remoteId, netAddress, port);
		linkedSystems.put(remoteId, updatedRemoteSys);
		logger.info("Actor System " + id + ". Remote System " + remoteId + " linked");
		return updatedRemoteSys;
	}

	/**
	 * Link a remote ActorSystem to this. If an ActorSystem with the same id is
	 * already linked, disconnection and reset with the new parameters are
	 * performed. Linking is lazy in that a connection will established only
	 * when the first actor-tell or summon is performed.
	 * 
	 * @param remoteId
	 *            the id of the remote ActorSystem
	 * @param netAddress
	 *            the IP address of the remote ActorSystem
	 * @param port
	 *            the port in which the ActorSystem listens to
	 */
	public void link(String remoteId, InetAddress netAddress, int port) {
		linkAndGet(remoteId, netAddress, port);
	}

	/**
	 * Calls {@link #dissolve(Actor actor, String reason)} without a reason.
	 * 
	 * @param actor
	 *            actor to dissolve
	 */
	public void dissolve(Actor actor) {
		dissolve(actor, "");
	}

	/**
	 * Request a stop for the actor and unregister it from this ActorSystem.
	 * 
	 * @param actor
	 *            actor to dissolve
	 * @param reason
	 *            the reason for dissolving
	 */
	public void dissolve(Actor actor, String reason) {
		String systemId = actor.address.systemId;
		if (systemId.equals(id)) {
			if (logger.isLoggable(Level.FINE)) {
				logger.fine("Actor System " + id + ". Dissolving local Actor " + actor.address.toString() + " : " + reason);
			}
			if (actor.isOperative()) {
				systemActor.tell(actor, Stop.create(reason));
			}
		} else {
			RemoteSystem remoteSystem = linkedSystems.get(systemId);
			if (remoteSystem == null) {
				String systemNotLinkedMessage = "Actor System " + id + ". System " + systemId + " is not linked";
				logger.severe(systemNotLinkedMessage);
				throw new IllegalArgumentException(systemNotLinkedMessage);
			}
			remoteSystem.dissolve(actor.address, reason);
		}
		actorsRegistry.remove(actor.address);
	}

	/**
	 * Calls {@link #dissolve(Actor actor, String reason)} without a reason.
	 * 
	 * @param address
	 *            the address of the actor to dissolve
	 */
	public void dissolve(String address) {
		dissolve(actorsRegistry.get(new ActorAddress(address)));
	}

	/**
	 * Calls {@link #dissolve(Actor actor, String reason)}
	 * 
	 * @param address
	 *            address of the actor to dissolve
	 * @param reason
	 *            the reason for dissolving
	 */
	public void dissolve(String address, String reason) {
		dissolve(actorsRegistry.get(new ActorAddress(address)), reason);
	}

	/**
	 * Calls {@link #shutdown(String)} with the reason from t
	 * 
	 * @param t
	 *            the throwable
	 */
	public void shutdown(Throwable t) {
		shutdown(t.getMessage());
	}

	/**
	 * Close this ActorSystem, stopping all local actors and all TCP connections
	 * 
	 * @param reason
	 *            the reason for shutting down
	 */
	public void shutdown(String reason) {
		systemActor.tell(systemActor, Stop.create(reason));
	}

	void tcpServerError(IOException e) {
		shutdown(e);
	}

	/**
	 * Change the Log level for this ActorSystem. This call is thread-safe.
	 * 
	 * @param newLevel
	 *            the new logging level
	 */
	public void setLogLevel(Level newLevel) {
		Handler[] handlers = this.logger.getHandlers();
		this.logger.setLevel(newLevel);
		for (Handler h : handlers) {
			h.setLevel(newLevel);
		}
	}

	protected void closeResources() {
		if (tcpListener != null) {
			tcpListener.stop();
		}
	}

	protected Iterator<Actor> registeredActors() {
		return actorsRegistry.values().iterator();
	}

}
