/* Copyright (c) 2016 Raffaele Arecchi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package aktrs.rt;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.logging.Level;

import aktrs.messages.AliveMessage;
import aktrs.messages.DissolveActorMessage;
import aktrs.messages.RemoteMessage;
import aktrs.messages.RemoteResponse;
import aktrs.messages.RemoteResponse.Result;
import aktrs.messages.SummonActorMessage;

public class ClientRequestHandler implements Runnable {

	private Socket clientSocket;
	private ActorSystem system;
	private String clientDescription;

	protected ClientRequestHandler(Socket clientSocket, ActorSystem system) {
		this.clientSocket = clientSocket;
		this.system = system;
		this.clientDescription = clientSocket.toString();
	}

	public void run() {
		system.logger.info("Starting Handler for client " + clientDescription);
		ObjectInputStream input = null;
		ObjectOutputStream output = null;
		try {
			input = new ObjectInputStream(clientSocket.getInputStream());
			output = new ObjectOutputStream(clientSocket.getOutputStream());

			while (true) {
				Object message;
				RemoteResponse response;
				try {
					message = input.readObject();
					if (system.logger.isLoggable(Level.FINE)) {
						system.logger.fine("Received message " + message.toString() + " from client " + clientDescription);
					}
					if (message == null) {
						break;
					}
					if (message instanceof AliveMessage) {
						response = new RemoteResponse();
						response.result = Result.OK;
					} else if (message instanceof SummonActorMessage) {
						response = handle((SummonActorMessage) message);
					} else if (message instanceof DissolveActorMessage) {
						response = handle((DissolveActorMessage) message);
					} else if (message instanceof RemoteMessage) {
						response = handle((RemoteMessage) message);
					} else {
						response = new RemoteResponse();
						response.result = Result.ERROR;
						response.description = "Unknown message: " + message.toString();
					}
				} catch (ClassNotFoundException e) {
					system.logger.log(Level.SEVERE, e.getMessage(), e);
					response = new RemoteResponse();
					response.result = Result.ERROR;
					response.description = e.getMessage();
				}
				if (system.logger.isLoggable(Level.FINE)) {
					system.logger.fine("Responding " + response.toString() + " to client " + clientDescription);
				}
				output.writeObject(response);
				output.flush();
			}
		} catch (SocketTimeoutException e) {
			system.logger.fine("Socket timeout for client socket " + clientDescription);
		} catch (Exception e) {
			system.logger.log(Level.SEVERE, e.getMessage(), e);
		} finally {
			if (input != null) {
				try {
					input.close();
				} catch (IOException e) {
				}
			}
			if (output != null) {
				try {
					output.close();
				} catch (IOException e) {
				}
			}
			if (clientSocket != null) {
				try {
					clientSocket.close();
				} catch (IOException e) {
				}
			}
		}
		system.logger.info("Terminated Handler for client " + clientDescription);
	}

	private RemoteResponse handle(DissolveActorMessage message) {
		RemoteResponse response = new RemoteResponse();
		try {
			system.dissolve(message.address.toString(), message.reason);
			response.result = Result.OK;
			response.description = "Actor " + message.address + " dissolved";
		} catch (Exception e) {
			system.logger.log(Level.SEVERE, e.getMessage(), e);
			response.result = Result.ERROR;
			response.description = e.getMessage();
		}
		return response;
	}

	protected RemoteResponse handle(SummonActorMessage message) {
		RemoteResponse response = new RemoteResponse();
		try {
			Actor summonedActor = system.summon(message.address, message.klass, message.initInput);
			response.result = Result.OK;
			response.description = "Actor " + summonedActor.address.toString() + " created";
		} catch (Exception e) {
			system.logger.log(Level.SEVERE, e.getMessage(), e);
			response.result = Result.ERROR;
			response.description = e.getMessage();
		}
		return response;
	}

	protected RemoteResponse handle(RemoteMessage message) {
		RemoteResponse response = new RemoteResponse();
		try {
			Actor recipientActor = system.evoke(message.recipientAddress);
			if (recipientActor == null) {
				response.result = Result.ERROR;
				response.description = "Unknown actor: " + message.recipientAddress.toString();
				return response;
			}
			Actor senderActor = system.evoke(message.senderAddress);
			if (senderActor == null) {
				RemoteSystem remoteSystem = system.linkIfNotLinked(message.senderAddress.systemId, clientSocket.getInetAddress(), message.remoteReplyPort);
				senderActor = system.mirrorRemoteActor(message.senderAddress, message.senderClass, remoteSystem);
			}
			senderActor.tell(recipientActor, message.message);
			response.result = Result.OK;
			response.description = "Message successfully delivered to " + recipientActor.address.toString();
		} catch (Exception e) {
			system.logger.log(Level.SEVERE, e.getMessage(), e);
			response.result = Result.ERROR;
			response.description = e.getMessage();
		}
		return response;
	}
}