/* Copyright (c) 2016 Raffaele Arecchi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package aktrs.rt;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;

public class TcpListener implements Runnable {

	private static final int CLIENT_SOCKET_TIMEOUT = 30000;
	private static final int SERVER_SOCKET_TIMEOUT = 30000;
	// private static final int CLIENT_SOCKET_TIMEOUT = 600000;

	private final ActorSystem actorSystem;
	protected final int port;

	private ServerSocket serverSocket = null;
	private boolean stopped = false;
	private ExecutorService threadPool = Executors.newFixedThreadPool(10);

	protected TcpListener(ActorSystem actorSystem, int port) {
		this.actorSystem = actorSystem;
		this.port = port;
	}

	public void run() {
		actorSystem.logger.info("Starting Tcp listener for Actor System " + actorSystem.id + ". Listening port: " + port);
		try {
			serverSocket = new ServerSocket(port);
			serverSocket.setSoTimeout(SERVER_SOCKET_TIMEOUT);
		} catch (IOException ioe) {
			actorSystem.logger.log(Level.SEVERE, ioe.getMessage(), ioe);
			actorSystem.tcpServerError(ioe);
		}
		while (!isStopped()) {
			try {
				actorSystem.logger.fine("Accepting connections for Actor System " + actorSystem.id);
				Socket clientSocket = serverSocket.accept();
				clientSocket.setSoTimeout(CLIENT_SOCKET_TIMEOUT);
				actorSystem.logger.info("Received a new connection request from " + clientSocket.toString());
				threadPool.execute(new ClientRequestHandler(clientSocket, actorSystem));
			} catch (SocketTimeoutException e) {
				actorSystem.logger.fine("Server socket timeout for Actor System " + actorSystem.id);
			} catch (Exception e) {
				actorSystem.logger.log(Level.SEVERE, e.getMessage(), e);
				reset();
			}
		}
		actorSystem.logger.info("Terminating Tcp listener for Actor System " + actorSystem.id);
		try {
			if (serverSocket != null) {
				serverSocket.close();
			}
		} catch (IOException ioe) {
		}
		threadPool.shutdown();
		actorSystem.logger.info("Tcp listener for Actor System " + actorSystem.id + " terminated");
	}

	private void reset() {
		actorSystem.logger.info("Resetting Tcp listener for Actor System " + actorSystem.id);
		try {
			if (serverSocket != null) {
				serverSocket.close();
			}
		} catch (IOException ioe) {
		}
		try {
			serverSocket = new ServerSocket(port);
		} catch (IOException e) {
			actorSystem.logger.log(Level.SEVERE, e.getMessage(), e);
			actorSystem.tcpServerError(e);
		}
		actorSystem.logger.info("Resetted Tcp listener for Actor System " + actorSystem.id);
	}

	private synchronized boolean isStopped() {
		return stopped;
	}

	public synchronized void stop() {
		actorSystem.logger.info("Stopping Tcp listener for Actor System " + actorSystem.id);
		this.stopped = true;
	}

}